#!/usr/bin/env python3
#
# Copyright 2018 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Project description variables"

PROJECT_NAME = 'kapitan'
VERSION = '0.19.0'
DESCRIPTION = ('Generic templated configuration management for Kubernetes, '
               'Terraform and other things')
AUTHOR = 'Ricardo Amaro'
AUTHOR_EMAIL = 'ramaro@google.com'
LICENCE = 'Apache License 2.0'
URL = 'https://github.com/deepmind/kapitan'
